
-- is the player in edit mode?
yl_speak_up.in_edit_mode = function(pname)
	return pname
		and yl_speak_up.edit_mode[pname]
		and (yl_speak_up.edit_mode[pname] == yl_speak_up.speak_to[pname].n_id)
end


-- in edit mode, *all* options are displayed
local old_calculate_displayable_options = yl_speak_up.calculate_displayable_options
yl_speak_up.calculate_displayable_options = function(pname, d_options, allow_recursion)
	-- no options - nothing to do
	if(not(d_options)) then
		return {}
	end
	if(yl_speak_up.in_edit_mode(pname)) then
		-- in edit mode: show all options (but sort them first)
		local retval = {}
		local sorted_list = yl_speak_up.get_sorted_options(d_options, "o_sort")
		for i, o_k in ipairs(sorted_list) do
			retval[o_k] = true
		end
		return retval
	end
	-- outside edit mode: really calculate what can be displayed
	return old_calculate_displayable_options(pname, d_options, allow_recursion)
end


-- in edit mode, autoanswer, random dialogs and d_got_item play no role and are *not* applied
-- (we want to see and edit all options regardless of preconditions)
local old_apply_autoanswer_etc = yl_speak_up.apply_autoanswer_and_random_and_d_got_item
yl_speak_up.apply_autoanswer_and_random_and_d_got_item = function(player, pname, d_id, dialog, allowed, active_dialog, recursion_depth)
	-- no automatic switching in edit_mode
	if(yl_speak_up.in_edit_mode(pname)) then
		return
	end
	return old_apply_autoanswer_etc(player, pname, d_id, dialog, allowed, active_dialog, recursion_depth)
end


-- helper function for yl_speak_up.get_fs_talkdialog:
--   shows the text the NPC "speaks" and adds edit and navigation buttons
--   (all only in *edit_mode*)
local old_talkdialog_main_text = yl_speak_up.get_fs_talkdialog_main_text
yl_speak_up.get_fs_talkdialog_main_text = function(pname, formspec, h, dialog, dialog_list, c_d_id, active_dialog)
	if(not(yl_speak_up.in_edit_mode(pname))) then
		return old_talkdialog_main_text(pname, formspec, h, dialog, dialog_list, c_d_id, active_dialog)
	end
	local d_id_to_dropdown_index = {}
	-- allow to change skin, wielded items etc.
	table.insert(formspec, "button[15.75,3.5;3.5,0.9;edit_skin;Edit Skin]")

	if(not(dialog) or not(dialog.n_dialogs)) then
		return {h = h, formspec = formspec, d_id_to_dropdown_index = {}, dialog_list = dialog_list}
	end

	-- display the window with the text the NPC is saying
	-- sort all dialogs by d_sort
	local sorted_list = yl_speak_up.get_sorted_options(dialog.n_dialogs, "d_sort")
	-- add buttons for previous/next dialog
	for i, d in ipairs(sorted_list) do
		-- build the list of available dialogs for the dropdown list(s)
		dialog_list = dialog_list..","..minetest.formspec_escape(d)
		if(d == c_d_id) then
			local prev_dialog = tostring(minetest.formspec_escape(sorted_list[i-1]))
			yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
				"button", "8.5,4.0;2,0.9", "prev_dialog_"..prev_dialog,
				"<",
				"Go to previous dialog "..prev_dialog..".",
				(sorted_list[ i-1 ]))
			local next_dialog = tostring(minetest.formspec_escape(sorted_list[i+1]))
			yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
				"button", "11,4.0;2,0.9", "next_dialog_"..next_dialog,
				">",
				"Go to next dialog "..next_dialog..".",
				(sorted_list[ i+1 ]))
		end
		d_id_to_dropdown_index[d] = i + 1
	end
	dialog_list = dialog_list..",d_end"
	d_id_to_dropdown_index["d_end"] = #sorted_list + 2

	table.insert(formspec, "label[0.2,4.6;Dialog:]") -- "..minetest.formspec_escape(c_d_id)..":]")
	table.insert(formspec, "dropdown[3.0,4.0;5,1;d_id;"..dialog_list..";"..
				(d_id_to_dropdown_index[c_d_id] or "1")..",]")
	table.insert(formspec, "tooltip[3.0,4.0;5,1;"..
		"Select the dialog you want to edit. Currently, dialog "..c_d_id..
			" is beeing displayed.;#FFFFFF;#000000]")

	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", "13.9,4.0;1,0.9", "show_new_dialog",
		"+",
		"Create a new dialog.",
		true)
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", "11.0,0.3;2,1.0", "button_edit_notes",
		"Notes",
		"Keep notes of what this NPC is for, how his character is etc.",
		true)
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", "13.2,0.3;2,0.9", "button_edit_name_and_description",
		"Edit",
		"Edit name and description of your NPC.",
		true)
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", "15.4,0.3;2,0.9", "button_save_dialog",
		"Save",
		"Save this dialog.",
		true)
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", "17.5,0.3;2.4,0.9", "button_export_dialog",
		"Export",
		"Export: Show the dialog in .json format which you can copy and store on your computer.",
		true)


	-- static help text instead of text input field for d_got_item
	if(c_d_id == "d_got_item") then
		table.insert(formspec, "hypertext[0.2,5;19.6,17.8;d_text;"..
			"<normal>Note:\nThis is a special dialog."..
				"It will be called when the player clicks on "..
				"<b>I want to give you something</b>."..
			"\nMost of the things listed below will be added automaticly when you add a "..
				"new option to this dialog. In most cases you may just have to edit the "..
				"<b>precondition</b> so that the <i>right item</i> is accepted, and then "..
				"set the <b>target dialog</b> <i>according to your needs</i>. Please also "..
				"edit the <b>alternate text</b> so that it fits your <i>item</i>!"..
			"\nThis is how it works in detail:"..
			"\n<b>Each option</b> you add here ought to deal with one item(stack) that "..
				"the NPC expects from the player, i.e. <i>farming:bread 2</i>. "..
				"Each option needs to be selected <b>automaticly</b> and ought to contain:"..
			"\n* a <b>precondition</b> regarding "..
				"<b>an item the player offered/gave to the NPC</b> "..
				"(shown as <b>player_offered_item</b> in overview) "..
				"where you define which item(stack) is relevant for this option"..
			"\n* an <b>effect</b> regarding <b>an item the player offered to the NPC</b> "..
				"(shown as <b>deal_with_offered_item</b> in overview) "..
				"where you define what shall happen to the offered item. Usually "..
				"the NPC will accept the item and put it into its inventory."..
			"\n* Don't forget to set a suitable target dialog for the <b>effect</b>! "..
				"Your NPC ought to comment on what he got, i.e. "..
				"<i>Thank you for those two breads! You saved me from starving.</i>"..
				"You can also work with an alternate text here (as is done in the "..
				"default setup when adding a new option here)."..
			"\n</normal>]")
	-- static help text instead of text input field for d_trade
	elseif(c_d_id == "d_trade") then
		table.insert(formspec, "hypertext[0.2,5;19.6,17.8;d_text;"..
			"<normal>Note:\nThis is a special dialog."..
				"It will be called when the player clicks on "..
				"<b>Let's trade!</b>."..
			"\nSome of the things listed below will be added automaticly when you add a "..
				"new option to this dialog. In most cases you may just have to edit the "..
				"<b>precondition</b> so that the <i>right item(stack)</i> is beeing "..
				"searched for, and you need to add suitable effects. The ones added "..
				"automaticly are just an example."..
			"\nNote that once the NPC found a matching precondition, it will execute the "..
			"relevant effects and present the player the trade list. Any further options "..
			"that might also fit will not be executed this time. Only <b>one</b> option "..
			"(or none) will be selected each time."..
			"\nThis is how it works in detail:"..
			"\n<b>Each option</b> you add here ought to deal with one item(stack) that "..
				"the NPC might or might not have in its inventory, "..
				"i.e. <i>default:stick 4</i>. "..
				"Each option needs to be selected <b>automaticly</b> and ought to contain:"..
			"\n* at least one <b>precondition</b> regarding "..
				"<b>the inventory of the NPC</b> "..
				"where you define which item(stack) is relevant for this option "..
				"(you can add multiple such preconditions for each option)"..
			"\n* at least one <b>effect</b> regarding what the NPC shall do if the "..
				"precondition matches. In most cases, <b>NPC crafts something</b>, "..
				"<b>put item from the NPC's inventory into a chest etc.</b> or "..
				"<b>take item from a chest etc. and put it into the NPC's inventory</b> "..
				"will be what you are looking for. More than one effect is possible."..
			"\n* In this particular case, no target dialog needs to be selected. After "..
				"executing the effect(s), the trade list view will be shown to the "..
				"player."..
			"\n</normal>]")
	elseif(active_dialog and active_dialog.d_text) then
		table.insert(formspec, "textarea[0.2,5;19.6,17.8;d_text;;"..
			minetest.formspec_escape(active_dialog.d_text or "?")..
			"]")
	else
		table.insert(formspec, "textarea[0.2,5;19.6,17.8;d_text;;"..
			minetest.formspec_escape("[no text]")..
			"]")
	end
	return {h = h, formspec = formspec, d_id_to_dropdown_index = d_id_to_dropdown_index,
		dialog_list = dialog_list}
end


-- helper function for yl_speak_up.get_fs_talkdialog:
--   prints one entry (option/answer) in normal and *edit_mode*
local old_talkdialog_line = yl_speak_up.get_fs_talkdialog_line
yl_speak_up.get_fs_talkdialog_line = function(
				formspec, h, pname_for_old_fs, oid, sb_v,
				dialog, allowed, pname,
				-- these additional parameters are needed *here*, in edit_mode:
				active_dialog, dialog_list, d_id_to_dropdown_index,
				current_index, anz_options)

	if(not(yl_speak_up.in_edit_mode(pname))) then
		-- in normal mode:
		return old_talkdialog_line(formspec, h, pname_for_old_fs, oid, sb_v,
				dialog, allowed, pname,
				active_dialog, dialog_list, d_id_to_dropdown_index,
				current_index, anz_options)
	end

	-- in edit mode:
	local offset = 0.0
	local field_length = 43.8
	if(pname_for_old_fs) then
		offset = 0.7
		field_length = 41.8
	end
	h = h + 1
	-- add a button "o_<nr>:" that leads to an edit formspec for this option
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", tostring(2.9+offset).."," .. h .. ";2,0.9", "edit_option_" .. oid,
		oid,
		"Edit target dialog, pre(C)onditions and (Ef)fects for option "..oid..".",
		true)
	-- find the right target dialog for this option (if it exists):
	local target_dialog = nil
	local results = active_dialog.d_options[sb_v.o_id].o_results
	-- has this option more results/effects than just switching to another dialog?
	local has_other_results = false
	if(results ~= nil) then
		for k, v in pairs(results) do
			if v.r_type == "dialog"
			  and (dialog.n_dialogs[v.r_value] ~= nil
			  or v.r_value == "d_end"
			  or v.r_value == "d_trade"
			  or v.r_value == "d_got_item") then
				-- there may be more than one in the data structure
				target_dialog = v.r_value
			elseif v.r_type ~= "dialog" then
				has_other_results = true
			end
		end
	end
	-- add a button "-> d_<nr>" that leads to the target dialog (if one is set)
	-- selecting an option this way MUST NOT execute the pre(C)onditions or (Ef)fects!
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", tostring(9.6+offset)..","..h..";1,0.9", "button_" .. oid,
		"->",
		"Go to target dialog "..minetest.formspec_escape(target_dialog or "")..
			" that will be shown when this option ("..oid..") is selected.",
		(target_dialog))

	-- allow to set a new target dialog
	table.insert(formspec, "dropdown["..tostring(5.0+offset)..","..h..";4.7,1;d_id_"..
		oid..";"..
		dialog_list..";"..
		(d_id_to_dropdown_index[(target_dialog or "?")] or "0")..",]")
	-- add a tooltip "Change target dialog"
	table.insert(formspec, "tooltip[5.0,"..h..";4.7,1;"..
		"Change target dialog for option "..oid..".;#FFFFFF;#000000]")

	-- are there any prerequirements?
	local prereq = active_dialog.d_options[sb_v.o_id].o_prerequisites
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", tostring(0.5+offset)..","..h..";0.5,0.9", "conditions_"..oid,
		"C",
		"There are pre(C)onditions required for showing this option. Display them.",
		(prereq and next(prereq)))

	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", tostring(1.6+offset)..","..h..";0.6,0.9", "effects_"..oid,
		"Ef",
		"There are further (Ef)fects (apart from switching\n"..
			"to a new dialog) set for this option. Display them.",
		(has_other_results))

	local d_option = active_dialog.d_options[sb_v.o_id]
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", tostring(2.25+offset)..","..h..";0.6,0.9", "quests_"..oid,
		"Q",
		"This option sets a (Q)est step if possible.\n"..
			"A special precondition is evaluated automaticly\n"..
			"to check if the quest step can be set.",
		(d_option and d_option.quest_id and d_option.quest_step))

	-- are there any actions defined?
	local actions = active_dialog.d_options[sb_v.o_id].actions
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", tostring(1.05+offset)..","..h..";0.5,0.9", "actions_"..oid,
		"A",
		"There is an (A)ction (i.e. a trade) that will happen\n"..
			"when switching to a new dialog. Display actions and\n"..
			"trade of this option.",
		(actions and next(actions)))

	if(sb_v.o_autoanswer) then
		table.insert(formspec,
			"label["..tostring(10.5+offset+0.2)..","..tostring(h+0.5)..";"..
				minetest.formspec_escape("[Automaticly selected if preconditions are met] "..
					tostring(sb_v.o_text_when_prerequisites_met))..
				"]")
	elseif(active_dialog.o_random) then
		table.insert(formspec,
			"label["..tostring(10.5+offset+0.2)..","..tostring(h+0.5)..";"..
				minetest.formspec_escape("[One of these options is randomly selected] "..
					tostring(sb_v.o_text_when_prerequisites_met))..
				"]")
	else
		-- show the actual text for the option
		yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
			"field", tostring(10.5+offset)..","..h..";"..
				 tostring(field_length-2.3)..",0.9",
			"text_option_" .. oid,
			";"..minetest.formspec_escape(sb_v.o_text_when_prerequisites_met),
			"Edit the text that is displayed on button "..oid..".",
			true)
	end
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
		"button", tostring(10.5+offset+field_length-2.2)..","..h..";1.0,0.9", "delete_option_"..oid,
		"Del",
		"Delete this option/answer.",
		true)

	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
--		"image_button", tostring(9.9+offset+field_length-0.5)..","..h..";0.5,0.9"..
--			";gui_furnace_arrow_bg.png^[transformR180",
		"button", tostring(10.5+offset+field_length-1.1)..","..h..";0.5,0.9",
		"option_move_down_"..oid,
		"v",
		"Move this option/answer one down.",
		(current_index < anz_options))
	yl_speak_up.add_formspec_element_with_tooltip_if(formspec,
--		"image_button", tostring(9.9+offset+field_length-1.0)..","..h..";0.5,0.9"..
--			";gui_furnace_arrow_bg.png",
		"button", tostring(10.5+offset+field_length-0.5)..","..h..";0.5,0.9",
		"option_move_up_"..oid,
		"^",
		"Move this option/answer one up.",
		(current_index > 1))
	return {h = h, formspec = formspec}
end
